#include "RHDPDADeterminizer.h"

namespace determinization {


void RHDPDADeterminizer::divideTransitions(const set<TransitionPDA>& transitions, set<TransitionPDA>& internalTransitions,
        set<TransitionPDA>& pushTransitions, set<TransitionPDA>& popTransitions) {
    for (set<TransitionPDA>::iterator transition = transitions.begin(); transition != transitions.end(); transition++) {
        if (transition->getPush().size() == 2) {
            pushTransitions.insert(*transition);
        } else if (transition->getPush().size() == 1) {
            internalTransitions.insert(*transition);
        } else {
            popTransitions.insert(*transition);
        }
    }
}


PDA RHDPDADeterminizer::determinize(PDA& rhdpda) {
    PDA rdpda;
    map<string, StateData> states;
    const set<TransitionPDA>& transitions = rhdpda.getTransitions();
    set<TransitionPDA> internalTransitions;
    set<TransitionPDA> pushTransitions;
    set<TransitionPDA> popTransitions;
    set<StateSymbolPair> allStateSymbolPairs = RHDPDADeterminizationUtils::buildAllStateSymbolPairs(rhdpda);
    vector<SComponent> allScomponents = RHDPDADeterminizationUtils::generateAllPossibleSComponents(allStateSymbolPairs);
    vector<RComponent> allRcomponents = RHDPDADeterminizationUtils::generateAllPossibleRComponents(allStateSymbolPairs);
    const set<Symbol>& inputAlphabet = rhdpda.getInputAlphabet();
    
    DeterminizationUtils::copyInputAlphabet(rhdpda, rdpda);
    divideTransitions(transitions, internalTransitions, pushTransitions, popTransitions);
    
    set<StateSymbolPair> initialSPairs = RHDPDADeterminizationUtils::combineStatesWithSymbol(rhdpda.getStates(),
            RHDPDADeterminizationUtils::BOTTOM_OF_STACK_SYMBOL);
    set<StateSymbolPair> initialRPairs = RHDPDADeterminizationUtils::combineStatesWithSymbol(rhdpda.getInitialStates(),
            RHDPDADeterminizationUtils::BOTTOM_OF_STACK_SYMBOL);
    const SComponent& initialS = RHDPDADeterminizationUtils::getSComponentWithPairsIdentity(initialSPairs);
    const RComponent& initialR(initialRPairs);
    const State& initialState = RHDPDADeterminizationUtils::getOrCreateState(initialS, initialR, states, rdpda);
    rdpda.addInitialState(initialState);

    while (true) {
        StateData* unmarkedStateData = RHDPDADeterminizationUtils::getUnmarkedState(states);
        if (unmarkedStateData == NULL) {
            break;
        }
        
        // Internal
        for (set<Symbol>::iterator a = inputAlphabet.begin(); a != inputAlphabet.end(); a++) {
            SComponent sPrimed;
            SComponent& s = unmarkedStateData->s;
            for (set<PairOfPairs>::iterator pairOfPairs = s.pairsOfPairs.begin();
                    pairOfPairs != s.pairsOfPairs.end();
                    pairOfPairs++) {
                const StateSymbolPair& x = pairOfPairs->pair1;
                const State& r = pairOfPairs->pair2.state;
                const Symbol& Y = pairOfPairs->pair2.symbol;
                for (set<TransitionPDA>::iterator transition = internalTransitions.begin();
                        transition != internalTransitions.end();
                        transition++) {
                    if (transition->getFrom() == r && transition->getInput() == *a && transition->getPop().front() == Y) {
                        const State& q = transition->getTo();
                        sPrimed.pairsOfPairs.insert(PairOfPairs(x, StateSymbolPair(q, Y)));
                    }
                }
            }
            RComponent rPrimed;
            RComponent& r = unmarkedStateData->r;
            for (set<StateSymbolPair>::iterator pair = r.pairs.begin(); pair != r.pairs.end(); pair++) {
                const State& p = pair->state;
                const Symbol& Y = pair->symbol;
                for (set<TransitionPDA>::iterator transition = internalTransitions.begin();
                        transition != internalTransitions.end();
                        transition++) {
                    if (transition->getFrom() == p && transition->getInput() == *a && transition->getPop().front() == Y) {
                        const State& q = transition->getTo();
                        rPrimed.pairs.insert(StateSymbolPair(q, Y));
                    }
                }
            }
            if (sPrimed.pairsOfPairs.size() > 0 && rPrimed.pairs.size() > 0) {
                const State& fromState = unmarkedStateData->state;
                const State& toState = RHDPDADeterminizationUtils::getOrCreateState(sPrimed, rPrimed, states, rdpda);
                const TransitionPDA transition(fromState, *a, toState);
                rdpda.addTransition(transition);
            }
        }
        
        // Push
        for (set<Symbol>::iterator a = inputAlphabet.begin(); a != inputAlphabet.end(); a++) {
            SComponent sDoublePrimed = RHDPDADeterminizationUtils::getSComponentWithPairsIdentity(allStateSymbolPairs);
            SComponent& s = unmarkedStateData->s;
            RComponent rDoublePrimed;
            RComponent& r = unmarkedStateData->r;
            for (set<StateSymbolPair>::iterator pair = r.pairs.begin(); pair != r.pairs.end(); pair++) {
                const State& p = pair->state;
                const Symbol& Y = pair->symbol;
                for (set<TransitionPDA>::iterator transition = pushTransitions.begin();
                        transition != pushTransitions.end();
                        transition++) {
                    if (transition->getFrom() == p && transition->getInput() == *a && transition->getPop().front() == Y) {
                        const State& q = transition->getTo();
                        const Symbol& Z = transition->getPush().front();
                        rDoublePrimed.pairs.insert(StateSymbolPair(q, Z));
                    }
                }
            }
            if (rDoublePrimed.pairs.size() > 0) {
                const State& fromState = unmarkedStateData->state;
                const State& toState =
                        RHDPDADeterminizationUtils::getOrCreateState(sDoublePrimed, rDoublePrimed, states, rdpda);
                string stackSymbolName = RHDPDADeterminizationUtils::buildStackSymbolName(s, r, *a, rdpda);
                list<Symbol> pop;
                list<Symbol> push(1, Symbol(stackSymbolName));
                const TransitionPDA transition(fromState, *a, toState, pop, push);
                rdpda.addTransition(transition);
            }
        }

        // Pop
        for (set<Symbol>::iterator a = inputAlphabet.begin(); a != inputAlphabet.end(); a++) {
            for (set<Symbol>::iterator b = inputAlphabet.begin(); b != inputAlphabet.end(); b++) {
                SComponent& s = unmarkedStateData->s;

                set<PairOfPairs> update;
                for (set<PairOfPairs>::iterator pairOfPairs = s.pairsOfPairs.begin();
                        pairOfPairs != s.pairsOfPairs.end();
                        pairOfPairs++) {
                    const State& r = pairOfPairs->pair1.state;
                    const Symbol& Z = pairOfPairs->pair1.symbol;
                    const State& rPrimed = pairOfPairs->pair2.state;
                    if (pairOfPairs->pair2.symbol == Z) {
                        for (set<TransitionPDA>::iterator pushTransition = pushTransitions.begin();
                                pushTransition != pushTransitions.end();
                                pushTransition++) {
                            for (set<TransitionPDA>::iterator popTransition = popTransitions.begin();
                                    popTransition != popTransitions.end();
                                    popTransition++) {
                                const State& p = pushTransition->getFrom();
                                const State& q = popTransition->getTo();
                                if (pushTransition->getInput() == *a && pushTransition->getTo() == r &&
                                        popTransition->getInput() == *b && popTransition->getFrom() == rPrimed &&
                                        pushTransition->getPush().front() == Z && popTransition->getPop().front() == Z) {
                                    const Symbol& Y = pushTransition->getPop().front();
                                    StateSymbolPair pair1(p, Y);
                                    StateSymbolPair pair2(q, Y);
                                    update.insert(PairOfPairs(pair1, pair2));
                                }
                            }
                        }
                    }
                }
                if (update.size() == 0) {
                    continue;
                }

                for (vector<SComponent>::iterator sPrimed = allScomponents.begin();
                        sPrimed != allScomponents.end();
                        sPrimed++) {
                    for (vector<RComponent>::iterator rPrimed = allRcomponents.begin();
                            rPrimed != allRcomponents.end();
                            rPrimed++) {
                        RComponent rDoublePrimed;
                        for (set<PairOfPairs>::iterator updateItem = update.begin();
                                updateItem != update.end();
                                updateItem++) {
                            const StateSymbolPair& y = updateItem->pair1;
                            const StateSymbolPair& x = updateItem->pair2;
                            if (rPrimed->pairs.find(y) != rPrimed->pairs.end()) {
                                rDoublePrimed.pairs.insert(x);
                            }
                        }
                        SComponent sDoublePrimed;
                        for (set<PairOfPairs>::iterator updateItem = update.begin();
                                updateItem != update.end();
                                updateItem++) {
                            const StateSymbolPair& z = updateItem->pair1;
                            const StateSymbolPair& y = updateItem->pair2;
                            for (set<PairOfPairs>::iterator sPrimedPair = sPrimed->pairsOfPairs.begin();
                                    sPrimedPair != sPrimed->pairsOfPairs.end();
                                    sPrimedPair++) {
                                if (sPrimedPair->pair2 == z) {
                                    const StateSymbolPair& x = sPrimedPair->pair1;
                                    sDoublePrimed.pairsOfPairs.insert(PairOfPairs(x, y));
                                }
                            }
                        }
                        if (sDoublePrimed.pairsOfPairs.size() > 0 && rDoublePrimed.pairs.size() > 0) {
                            const State& fromState = unmarkedStateData->state;
                            const State& toState =
                                    RHDPDADeterminizationUtils::getOrCreateState(sDoublePrimed, rDoublePrimed, states, rdpda);
                            string stackSymbolName =
                                    RHDPDADeterminizationUtils::buildStackSymbolName(*sPrimed, *rPrimed, *a, rdpda);
                            list<Symbol> pop(1, Symbol(stackSymbolName));
                            list<Symbol> push;
                            const TransitionPDA transition(fromState, *b, toState, pop, push);
                            rdpda.addTransition(transition);
                        }
                    }
                }
            }
        }
    }

    const set<State>& rhdpdaFinalState = rhdpda.getFinalStates();
    for (map<string, StateData>::iterator state = states.begin();
            state != states.end();
            state++) {
        bool isFinalState = false;
        RComponent& r = state->second.r;
        for (set<StateSymbolPair>::iterator pair = r.pairs.begin();
                pair != r.pairs.end();
                pair++) {
            if (rhdpdaFinalState.find(pair->state) != rhdpdaFinalState.end()) {
                isFinalState = true;
            }
        }
        if (isFinalState) {
            rdpda.addFinalState(state->second.state);
        }
    }

    return rdpda;
}

} /* namespace determinization */
