/*
 * RegExpParser.cpp
 *
 *  Created on: Nov 23, 2013
 *      Author: Martin Zak
 */

#include "RegExpParser.h"
#include "../sax/ParserException.h"

namespace regexp {

RegExp RegExpParser::parse(list<Token>& input) {
	popToken(input, Token::START_ELEMENT, "regexp");
	RegExp regexp;
	regexp.setRegExp(parseElement(input));
	popToken(input, Token::END_ELEMENT, "regexp");

	return regexp;
}

RegExpElement* RegExpParser::parseElement(list<Token>& input) {
	if (isToken(input, Token::START_ELEMENT, "symbol")) {
		return parseSymbol(input);
	} else if (isToken(input, Token::START_ELEMENT, "empty")) {
		return parseEmpty(input);
	} else if (isToken(input, Token::START_ELEMENT, "epsilon")) {
		return parseEpsilon(input);
	} else if (isToken(input, Token::START_ELEMENT, "iteration")) {
		return parseIteration(input);
	} else if (isToken(input, Token::START_ELEMENT, "alternation")) {
		return parseAlternation(input);
	} else if (isToken(input, Token::START_ELEMENT, "concatenation")) {
		return parseConcatenation(input);
	} else {
		return NULL;
	}
}

Alternation* RegExpParser::parseAlternation(list<Token>& input) {
	popToken(input, Token::START_ELEMENT, "alternation");

	Alternation* alternation = new Alternation;
	parseContent(input, alternation->getElements());

	popToken(input, Token::END_ELEMENT, "alternation");
	return alternation;
}

Concatenation* RegExpParser::parseConcatenation(list<Token>& input) {
	popToken(input, Token::START_ELEMENT, "concatenation");

	Concatenation* concatenation = new Concatenation();
	parseContent(input, concatenation->getElements());

	popToken(input, Token::END_ELEMENT, "concatenation");
	return concatenation;

}

Iteration* RegExpParser::parseIteration(list<Token>& input) {
	popToken(input, Token::START_ELEMENT, "iteration");

	Iteration* iteration = new Iteration();
	iteration->setElement(parseElement(input));

	popToken(input, Token::END_ELEMENT, "iteration");
	return iteration;
}

void RegExpParser::parseContent(list<Token>& input, list<RegExpElement*>& elements) {
	while (true) {
		RegExpElement* element = parseElement(input);
		if(!element) return;
		elements.push_back(element);
	}
}

RegExpEpsilon* RegExpParser::parseEpsilon(list<Token>& input) {
	popToken(input, Token::START_ELEMENT, "epsilon");

	RegExpEpsilon* epsilon = new RegExpEpsilon();

	popToken(input, Token::END_ELEMENT, "epsilon");
	return epsilon;
}

RegExpEmpty* RegExpParser::parseEmpty(list<Token>& input) {
	popToken(input, Token::START_ELEMENT, "empty");

	RegExpEmpty* empty = new RegExpEmpty();

	popToken(input, Token::END_ELEMENT, "empty");
	return empty;
}

RegExpSymbol* RegExpParser::parseSymbol(list<Token>& input) {
	popToken(input, Token::START_ELEMENT, "symbol");

	if (input.front().getType() == Token::CHARACTER) {
		RegExpSymbol* symbol = new RegExpSymbol(input.front().getData());
		input.pop_front();
		popToken(input, Token::END_ELEMENT, "symbol");
		return symbol;
	} else {
		throw ParserException(Token("", Token::CHARACTER), input.front());
	}
}

bool RegExpParser::isToken(list<Token>& input, Token::TokenType type, string data) {
	return input.front().getType() == type && input.front().getData() == data;
}

void RegExpParser::popToken(list<Token>& input, Token::TokenType type, string data) {
	if (isToken(input, type, data)) {
		input.pop_front();
	} else {
		throw ParserException(Token(data, type), input.front());
	}
}

} /* namespace regexp */

