/*
 * RegExpOptimize.cpp
 *
 *  Created on: 20. 1. 2014
 *      Author: tomas
 */

#include "RegExpOptimize.h"

using namespace alib;
using namespace regexp;

RegExp RegExpOptimize::optimize( const RegExp & regexp )
{
    RegExpElement* optimized = optimize( regexp.getRegExp( ) );
    RegExp ret;
    ret.setRegExp( optimized );

    return ret;
}

RegExpElement* RegExpOptimize::optimize( const RegExpElement * node )
{
    const Alternation * alternation = dynamic_cast<const Alternation*>( node );
    const Concatenation * concatenation = dynamic_cast<const Concatenation*>( node );
    const Iteration * iteration = dynamic_cast<const Iteration*>( node );
    const RegExpSymbol * symbol = dynamic_cast<const RegExpSymbol*>( node );
    const RegExpEmpty * empty= dynamic_cast<const RegExpEmpty*>( node );
    const RegExpEpsilon * eps = dynamic_cast<const RegExpEpsilon*>( node );

    if( alternation )
        return optimize( alternation );
    if( concatenation )
        return optimize( concatenation );
    if( iteration )
        return optimize( iteration );
    if( symbol )
        return optimize( symbol );
    if( empty )
        return optimize( empty );
    if( eps )
        return optimize( eps );

    throw AlibException( "RegExpOptimize::optimize - unknown RegExpElement node" );
}

RegExpElement * RegExpOptimize::optimize( const Alternation * node )
{
    Alternation* alt = new Alternation( );

    for( const auto & child : node->getElements( ) )
        alt->getElements( ).push_back( optimize( child ) );

    // Targets for optimization: Melichar, 2.87: Rule A1, A2, A3, A4, A8, A9

    // A1: x + ( y + z ) = ( x + y ) + z
    for( auto it = alt->getElements( ).begin( ) ; it != alt->getElements( ).end( ); )
    {
        Alternation * childAlt = dynamic_cast<Alternation*> ( * it );
        if( childAlt )
        {
            auto it2 = it;
            it2 ++;

            alt->getElements( ).insert( it2, childAlt->getElements( ).begin( ), childAlt->getElements( ).end( ) );

            childAlt->getElements( ).clear( );
            delete childAlt;
            it = alt->getElements( ).erase( it );
        }
        else
            it ++;
    }

    // A2: x + y = y + x
    alt->getElements().sort([](RegExpElement const * const & a, RegExpElement const * const & b) -> bool {
      return *a < *b;
    });

    // A3: x + EMPTY = x
    for( auto it = alt->getElements( ).begin( ) ; it != alt->getElements( ).end( ); )
    {
        if ( dynamic_cast<RegExpEmpty*>( *it ) )
        {
            delete *it;
            it = alt->getElements( ).erase( it );
        }
        else
            it ++;
    }

    // TODO: A4: x + x = x
        // depends on RegExpElement::operator<


    // Empty subtree and alternation of single node
    if( alt->getElements( ).size( ) == 0 )
    {
        delete alt;
        return new RegExpEmpty( );
    }
    else if( alt->getElements( ).size( ) == 1 )
    {
        RegExpElement * ret = * alt->getElements( ).begin( );
        alt->getElements( ).clear( );
        delete alt;
        return ret;
    }

    return alt;
}

RegExpElement * RegExpOptimize::optimize( const Concatenation * node )
{
    Concatenation* concat = new Concatenation( );

    for( const auto & child : node->getElements( ) )
        concat->getElements( ).push_back( optimize( child ) );

    // Targets for optimization: Melichar, 2.87: Rule A5, A6, A7

    // A5: a.(b.c) = (a.b).c
    for( auto it = concat->getElements( ).begin( ) ; it != concat->getElements( ).end( ) ; )
    {
        Concatenation * childConcat = dynamic_cast<Concatenation*>( * it );
        if( childConcat )
        {
            auto it2 = it;
            it2 ++;

            concat->getElements( ).insert( it2, childConcat->getElements( ).begin( ), childConcat->getElements( ).end( ) );

            childConcat->getElements( ).clear( );
            delete childConcat;
            it = concat->getElements( ).erase( it );
        }
        else
            it ++;
    }

    // A7: EMPTY.x = EMPTY
    for( const auto & childNode : concat->getElements( ) )
    {
        if( dynamic_cast<RegExpEmpty*>( childNode ) )
        {
            delete concat;
            return new RegExpEmpty( );
        }
    }

    // A6: EPS.x = x.EPS = x
    for( auto it = concat->getElements( ).begin( ) ; it != concat->getElements( ).end( ); )
    {
        // EPS.EPS.EPS.EPS = EPS
        if( dynamic_cast<RegExpEpsilon*>( *it ) && concat->getElements( ).size( ) > 1 )
        {
            delete * it;
            it = concat->getElements( ).erase( it );
        }
        else
            it ++;
    }

    if( concat->getElements( ).size( ) == 1 )
    {
        RegExpElement * ret = * concat->getElements( ).begin( );
        concat->getElements( ).clear( );
        delete concat;
        return ret;
    }

    return concat;
}

RegExpElement * RegExpOptimize::optimize( const Iteration * node )
{
    Iteration* iter = new Iteration( );
    iter->setElement( optimize( node->getElement( ) ) );

    // Targets for optimization: Melichar, 2.87: Rule A10, A11, a** = a*

    // a** = a*
    Iteration* childIteration;
    while( ( childIteration = dynamic_cast<Iteration*>( iter->getElement( ) ) ) )
    {
        iter->setElement( childIteration->getElement( ) );
        childIteration->setElement( NULL );
        delete childIteration;
    }

    // EMPTY* = eps
    if( dynamic_cast<RegExpEmpty*>( iter->getElement( ) ) )
    {
        delete iter;
        return new RegExpEpsilon( );
    }

    // TODO: A10: x* = EPS + x*x    // prob must be done in Concatenation node
    // TODO: A11: x* = ( EPS + x )*

    return iter;
}

RegExpElement * RegExpOptimize::optimize( const RegExpSymbol * node )
{
    return node->clone( );
}

RegExpElement * RegExpOptimize::optimize( const RegExpEmpty * node )
{
    return node->clone( );
}

RegExpElement * RegExpOptimize::optimize( const RegExpEpsilon * node )
{
    return node->clone( );
}
